@@ -19,7 +19,7 @@
 # FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 # AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 # LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-# OUT OF OR IN CONNECT<ION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
+# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 # THE SOFTWARE.
 #
 ###############################################################################
@@ -30,6 +30,9 @@ import six
 
 from autobahn.util import public
 
+from autobahn.wamp.request import Subscription, Registration
+
+
 __all__ = (
     'ComponentConfig',
     'HelloReturn',
@@ -46,6 +49,7 @@ __all__ = (
     'CallDetails',
     'CallOptions',
     'CallResult',
+    'EncodedPayload'
 )
 
 
@@ -107,7 +111,7 @@ class ComponentConfig(object):
         self.shared = shared
 
     def __str__(self):
-        return u"ComponentConfig(realm=<{0}>, extra={1}, keyring={2}, controller={3}, shared={4})".format(self.realm, self.extra, self.keyring, self.controller, self.shared)
+        return u"ComponentConfig(realm=<{}>, extra={}, keyring={}, controller={}, shared={})".format(self.realm, self.extra, self.keyring, self.controller, self.shared)
 
 
 @public
@@ -168,7 +172,7 @@ class Accept(HelloReturn):
         self.authextra = authextra
 
     def __str__(self):
-        return u"Accept(realm=<{0}>, authid=<{1}>, authrole=<{2}>, authmethod={3}, authprovider={4}, authextra={5})".format(self.realm, self.authid, self.authrole, self.authmethod, self.authprovider, self.authextra)
+        return u"Accept(realm=<{}>, authid=<{}>, authrole=<{}>, authmethod={}, authprovider={}, authextra={})".format(self.realm, self.authid, self.authrole, self.authmethod, self.authprovider, self.authextra)
 
 
 @public
@@ -198,7 +202,7 @@ class Deny(HelloReturn):
         self.message = message
 
     def __str__(self):
-        return u"Deny(reason=<{0}>, message='{1}')".format(self.reason, self.message)
+        return u"Deny(reason=<{}>, message='{}')".format(self.reason, self.message)
 
 
 @public
@@ -229,7 +233,7 @@ class Challenge(HelloReturn):
         self.extra = extra or {}
 
     def __str__(self):
-        return u"Challenge(method={0}, extra={1})".format(self.method, self.extra)
+        return u"Challenge(method={}, extra={})".format(self.method, self.extra)
 
 
 @public
@@ -406,7 +410,7 @@ class CloseDetails(object):
         self.message = message
 
     def __str__(self):
-        return u"CloseDetails(reason=<{0}>, message='{1}')".format(self.reason, self.message)
+        return u"CloseDetails(reason=<{}>, message='{}')".format(self.reason, self.message)
 
 
 @public
@@ -418,31 +422,47 @@ class SubscribeOptions(object):
 
     __slots__ = (
         'match',
+        'details',
         'details_arg',
         'get_retained',
+        'correlation',
     )
 
-    def __init__(self, match=None, details_arg=None, get_retained=None):
+    def __init__(self, match=None, details=None, details_arg=None, get_retained=None, correlation=None):
         """
 
         :param match: The topic matching method to be used for the subscription.
         :type match: str
 
-        :param details_arg: When invoking the handler, provide event details
-          in this keyword argument to the callable.
+        :param details: When invoking the handler, provide event details in a keyword
+            parameter ``details``.
+        :type details: bool
+
+        :param details_arg: DEPCREATED (use "details" flag). When invoking the handler
+            provide event details in this keyword argument to the callable.
         :type details_arg: str
 
         :param get_retained: Whether the client wants the retained message we may have along with the subscription.
         :type get_retained: bool or None
         """
         assert(match is None or (type(match) == six.text_type and match in [u'exact', u'prefix', u'wildcard']))
+        assert(details is None or (type(details) == bool and details_arg is None))
         assert(details_arg is None or type(details_arg) == str)  # yes, "str" is correct here, since this is about Python identifiers!
         assert(get_retained is None or type(get_retained) is bool)
 
         self.match = match
-        self.details_arg = details_arg
+        self.details = details
+
+        # FIXME: this is for backwards compat, but we'll deprecate it in the future
+        if details:
+            self.details_arg = 'details'
+        else:
+            self.details_arg = details_arg
+
         self.get_retained = get_retained
 
+        self.correlation = correlation
+
     def message_attr(self):
         """
         Returns options dict as sent within WAMP messages.
@@ -458,7 +478,7 @@ class SubscribeOptions(object):
         return options
 
     def __str__(self):
-        return u"SubscribeOptions(match={0}, details_arg={1}, get_retained={2})".format(self.match, self.details_arg, self.get_retained)
+        return u"SubscribeOptions(match={}, details={}, details_arg={}, get_retained={})".format(self.match, self.details, self.details_arg, self.get_retained)
 
 
 @public
@@ -469,6 +489,7 @@ class EventDetails(object):
     """
 
     __slots__ = (
+        'subscription',
         'publication',
         'publisher',
         'publisher_authid',
@@ -478,9 +499,12 @@ class EventDetails(object):
         'enc_algo',
     )
 
-    def __init__(self, publication, publisher=None, publisher_authid=None, publisher_authrole=None, topic=None, retained=None, enc_algo=None):
+    def __init__(self, subscription, publication, publisher=None, publisher_authid=None, publisher_authrole=None, topic=None, retained=None, enc_algo=None):
         """
 
+        :param subscription: The (client side) subscription object on which this event is delivered.
+        :type subscription: instance of :class:`autobahn.wamp.request.Subscription`
+
         :param publication: The publication ID of the event (always present).
         :type publication: int
 
@@ -490,31 +514,33 @@ class EventDetails(object):
 
         :param publisher_authid: The WAMP authid of the original publisher of this event.
             Only filled when publisher is disclosed.
-        :type publisher_authid: None or str
+        :type publisher_authid: str or None
 
         :param publisher_authrole: The WAMP authrole of the original publisher of this event.
             Only filled when publisher is disclosed.
-        :type publisher_authrole: None or str
+        :type publisher_authrole: str or None
 
         :param topic: For pattern-based subscriptions, the actual topic URI being published to.
             Only filled for pattern-based subscriptions.
-        :type topic: None or str
+        :type topic: str or None
 
         :param retained: Whether the message was retained by the broker on the topic, rather than just published.
         :type retained: bool or None
 
         :param enc_algo: Payload encryption algorithm that
             was in use (currently, either ``None`` or ``u'cryptobox'``).
-        :type enc_algo: None or str
+        :type enc_algo: str or None
         """
+        assert(isinstance(subscription, Subscription))
         assert(type(publication) in six.integer_types)
         assert(publisher is None or type(publisher) in six.integer_types)
         assert(publisher_authid is None or type(publisher_authid) == six.text_type)
         assert(publisher_authrole is None or type(publisher_authrole) == six.text_type)
         assert(topic is None or type(topic) == six.text_type)
         assert(retained is None or type(retained) is bool)
-        assert(enc_algo is None or (type(enc_algo) == six.text_type and enc_algo in [u'cryptobox']))
+        assert(enc_algo is None or type(enc_algo) == six.text_type)
 
+        self.subscription = subscription
         self.publication = publication
         self.publisher = publisher
         self.publisher_authid = publisher_authid
@@ -524,7 +550,7 @@ class EventDetails(object):
         self.enc_algo = enc_algo
 
     def __str__(self):
-        return u"EventDetails(publication={}, publisher={}, publisher_authid={}, publisher_authrole={}, topic=<{}>, retained={}, enc_algo={})".format(self.publication, self.publisher, self.publisher_authid, self.publisher_authrole, self.topic, self.retained, self.enc_algo)
+        return u"EventDetails(subscription={}, publication={}, publisher={}, publisher_authid={}, publisher_authrole={}, topic=<{}>, retained={}, enc_algo={})".format(self.subscription, self.publication, self.publisher, self.publisher_authid, self.publisher_authrole, self.topic, self.retained, self.enc_algo)
 
 
 @public
@@ -544,6 +570,7 @@ class PublishOptions(object):
         'eligible_authid',
         'eligible_authrole',
         'retain',
+        'correlation',
     )
 
     def __init__(self,
@@ -555,7 +582,8 @@ class PublishOptions(object):
                  eligible=None,
                  eligible_authid=None,
                  eligible_authrole=None,
-                 retain=None):
+                 retain=None,
+                 correlation=None):
         """
 
         :param acknowledge: If ``True``, acknowledge the publication with a success or
@@ -607,6 +635,8 @@ class PublishOptions(object):
         self.eligible_authrole = eligible_authrole
         self.retain = retain
 
+        self.correlation = correlation
+
     def message_attr(self):
         """
         Returns options dict as sent within WAMP messages.
@@ -643,7 +673,7 @@ class PublishOptions(object):
         return options
 
     def __str__(self):
-        return u"PublishOptions(acknowledge={0}, exclude_me={1}, exclude={2}, exclude_authid={3}, exclude_authrole={4}, eligible={5}, eligible_authid={6}, eligible_authrole={7}, retain={8})".format(self.acknowledge, self.exclude_me, self.exclude, self.exclude_authid, self.exclude_authrole, self.eligible, self.eligible_authid, self.eligible_authrole, self.retain)
+        return u"PublishOptions(acknowledge={}, exclude_me={}, exclude={}, exclude_authid={}, exclude_authrole={}, eligible={}, eligible_authid={}, eligible_authrole={}, retain={})".format(self.acknowledge, self.exclude_me, self.exclude, self.exclude_authid, self.exclude_authrole, self.eligible, self.eligible_authid, self.eligible_authrole, self.retain)
 
 
 @public
@@ -657,10 +687,13 @@ class RegisterOptions(object):
         'match',
         'invoke',
         'concurrency',
+        'force_reregister',
         'details_arg',
+        'correlation',
     )
 
-    def __init__(self, match=None, invoke=None, concurrency=None, details_arg=None):
+    def __init__(self, match=None, invoke=None, concurrency=None, details_arg=None,
+                 force_reregister=None, correlation=None):
         """
 
         :param details_arg: When invoking the endpoint, provide call details
@@ -671,11 +704,15 @@ class RegisterOptions(object):
         assert(invoke is None or (type(invoke) == six.text_type and invoke in [u'single', u'first', u'last', u'roundrobin', u'random']))
         assert(concurrency is None or (type(concurrency) in six.integer_types and concurrency > 0))
         assert(details_arg is None or type(details_arg) == str)  # yes, "str" is correct here, since this is about Python identifiers!
+        assert force_reregister in [None, True, False]
 
         self.match = match
         self.invoke = invoke
         self.concurrency = concurrency
         self.details_arg = details_arg
+        self.force_reregister = force_reregister
+
+        self.correlation = correlation
 
     def message_attr(self):
         """
@@ -692,10 +729,13 @@ class RegisterOptions(object):
         if self.concurrency is not None:
             options[u'concurrency'] = self.concurrency
 
+        if self.force_reregister is not None:
+            options[u'force_reregister'] = self.force_reregister
+
         return options
 
     def __str__(self):
-        return u"RegisterOptions(match={0}, invoke={1}, concurrency={2}, details_arg={3})".format(self.match, self.invoke, self.concurrency, self.details_arg)
+        return u"RegisterOptions(match={}, invoke={}, concurrency={}, details_arg={}, force_reregister={})".format(self.match, self.invoke, self.concurrency, self.details_arg, self.force_reregister)
 
 
 @public
@@ -706,6 +746,7 @@ class CallDetails(object):
     """
 
     __slots__ = (
+        'registration',
         'progress',
         'caller',
         'caller_authid',
@@ -714,38 +755,43 @@ class CallDetails(object):
         'enc_algo',
     )
 
-    def __init__(self, progress=None, caller=None, caller_authid=None, caller_authrole=None, procedure=None, enc_algo=None):
+    def __init__(self, registration, progress=None, caller=None, caller_authid=None, caller_authrole=None, procedure=None, enc_algo=None):
         """
 
+        :param registration: The (client side) registration object this invocation is delivered on.
+        :type registration: instance of :class:`autobahn.wamp.request.Registration`
+
         :param progress: A callable that will receive progressive call results.
-        :type progress: callable
+        :type progress: callable or None
 
         :param caller: The WAMP session ID of the caller, if the latter is disclosed.
             Only filled when caller is disclosed.
-        :type caller: int
+        :type caller: int or None
 
         :param caller_authid: The WAMP authid of the original caller of this event.
             Only filled when caller is disclosed.
-        :type caller_authid: None or str
+        :type caller_authid: str or None
 
         :param caller_authrole: The WAMP authrole of the original caller of this event.
             Only filled when caller is disclosed.
-        :type caller_authrole: None or str
+        :type caller_authrole: str or None
 
         :param procedure: For pattern-based registrations, the actual procedure URI being called.
-        :type procedure: None or str
+        :type procedure: str or None
 
         :param enc_algo: Payload encryption algorithm that
             was in use (currently, either `None` or `"cryptobox"`).
-        :type enc_algo: None or str
+        :type enc_algo: str or None
         """
+        assert(isinstance(registration, Registration))
         assert(progress is None or callable(progress))
         assert(caller is None or type(caller) in six.integer_types)
         assert(caller_authid is None or type(caller_authid) == six.text_type)
         assert(caller_authrole is None or type(caller_authrole) == six.text_type)
         assert(procedure is None or type(procedure) == six.text_type)
-        assert(enc_algo is None or (type(enc_algo) == six.text_type and enc_algo in [u'cryptobox']))
+        assert(enc_algo is None or type(enc_algo) == six.text_type)
 
+        self.registration = registration
         self.progress = progress
         self.caller = caller
         self.caller_authid = caller_authid
@@ -754,7 +800,7 @@ class CallDetails(object):
         self.enc_algo = enc_algo
 
     def __str__(self):
-        return u"CallDetails(progress={0}, caller={1}, caller_authid={2}, caller_authrole={3}, procedure=<{4}>, enc_algo={5})".format(self.progress, self.caller, self.caller_authid, self.caller_authrole, self.procedure, self.enc_algo)
+        return u"CallDetails(registration={}, progress={}, caller={}, caller_authid={}, caller_authrole={}, procedure=<{}>, enc_algo={})".format(self.registration, self.progress, self.caller, self.caller_authid, self.caller_authrole, self.procedure, self.enc_algo)
 
 
 @public
@@ -766,11 +812,13 @@ class CallOptions(object):
     __slots__ = (
         'on_progress',
         'timeout',
+        'correlation',
     )
 
     def __init__(self,
                  on_progress=None,
-                 timeout=None):
+                 timeout=None,
+                 correlation=None):
         """
 
         :param on_progress: A callback that will be called when the remote endpoint
@@ -786,6 +834,8 @@ class CallOptions(object):
         self.on_progress = on_progress
         self.timeout = timeout
 
+        self.correlation = correlation
+
     def message_attr(self):
         """
         Returns options dict as sent within WAMP messages.
@@ -801,7 +851,7 @@ class CallOptions(object):
         return options
 
     def __str__(self):
-        return u"CallOptions(on_progress={0}, timeout={1})".format(self.on_progress, self.timeout)
+        return u"CallOptions(on_progress={}, timeout={})".format(self.on_progress, self.timeout)
 
 
 @public
@@ -827,14 +877,53 @@ class CallResult(object):
         :type kwresults: dict
         """
         enc_algo = kwresults.pop('enc_algo', None)
-        assert(enc_algo is None or (type(enc_algo) == six.text_type and enc_algo in [u'cryptobox']))
+        assert(enc_algo is None or type(enc_algo) == six.text_type)
 
         self.enc_algo = enc_algo
         self.results = results
         self.kwresults = kwresults
 
     def __str__(self):
-        return u"CallResult(results={0}, kwresults={1}, enc_algo={2})".format(self.results, self.kwresults, self.enc_algo)
+        return u"CallResult(results={}, kwresults={}, enc_algo={})".format(self.results, self.kwresults, self.enc_algo)
+
+
+@public
+class EncodedPayload(object):
+    """
+    Wrapper holding an encoded application payload when using WAMP payload transparency.
+    """
+
+    __slots__ = (
+        'payload',
+        'enc_algo',
+        'enc_serializer',
+        'enc_key'
+    )
+
+    def __init__(self, payload, enc_algo, enc_serializer=None, enc_key=None):
+        """
+
+        :param payload: The encoded application payload.
+        :type payload: bytes
+
+        :param enc_algo: The payload transparency algorithm identifier to check.
+        :type enc_algo: str
+
+        :param enc_serializer: The payload transparency serializer identifier to check.
+        :type enc_serializer: str
+
+        :param enc_key: If using payload transparency with an encryption algorithm, the payload encryption key.
+        :type enc_key: str or None
+        """
+        assert(type(payload) == six.binary_type)
+        assert(type(enc_algo) == six.text_type)
+        assert(enc_serializer is None or type(enc_serializer) == six.text_type)
+        assert(enc_key is None or type(enc_key) == six.text_type)
+
+        self.payload = payload
+        self.enc_algo = enc_algo
+        self.enc_serializer = enc_serializer
+        self.enc_key = enc_key
 
 
 class IPublication(object):
